//////////////////////////////////////////////////
/////////////View Frustum Splitting///////////////
//////////////////////////////////////////////////
//////////////////////////////////////////////////
////24.12.2008////////////////by Nils Daumann/////
///////////////www.ssmasters.de///////////////////
//////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>

//engine specific header file
#define DLL_USE		//always define before including adll.h
#include "adll.h"
#include <d3dx9.h>

#define M_PI 1.5707963267948966192313216916397514L

//DLL main entry point - normally this needs not to be changed
BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved )
{
	//make sure the library is linked
	engine_bind();
	return TRUE;
}

//
DLLFUNC void SM_SplitMatrix_create(VIEW *CamView, VIEW *LightView, var clipnear, var clipfar)
{
	D3DXMATRIX matCamView;
	D3DXMATRIX matCamProj;
	D3DXMATRIX matCamViewProj;
	D3DXMATRIX matCamInvViewProj;
	D3DXVECTOR3 vecCamPos;
	D3DXVECTOR3 vecCamAngle;
	D3DXVECTOR3 vecCamLookAt;
	D3DXVECTOR3 vecCamUp;
	D3DXVECTOR3 vecCamDir;

	D3DXMATRIX matLightView;
	D3DXMATRIX matLightProj;
	D3DXMATRIX matLightViewProj;
	D3DXVECTOR3 vecLightPos;
	D3DXVECTOR3 vecLightAngle;
	D3DXVECTOR3 vecLightLookAt;
	D3DXVECTOR3 vecLightUp;
	D3DXVECTOR3 vecLightDir;

	VECTOR vecLookAtTemp;
	VECTOR vecUpTemp;
	ANGLE angDirTemp;

	//get the camera position
	vecCamPos = D3DXVECTOR3(_FLOAT(CamView->x), _FLOAT(CamView->z), _FLOAT(CamView->y));

	//create a temporare look at vector
	vecLookAtTemp.x = 10000.0;
	vecLookAtTemp.y = 0.0;
	vecLookAtTemp.z = 0.0;

	//get the camera direction
	angDirTemp.pan = CamView->pan;
	angDirTemp.tilt = CamView->tilt;
	angDirTemp.roll = CamView->roll;

	//rotate the cameras look at vector and set the final one
	vec_rotate(&vecLookAtTemp,&angDirTemp);
	vecCamLookAt = D3DXVECTOR3(_FLOAT(vecLookAtTemp.x),_FLOAT(vecLookAtTemp.z),_FLOAT(vecLookAtTemp.y));
	vecCamLookAt += vecCamPos;

	vecCamUp = D3DXVECTOR3(0,1,0);

	//calculate the cameras view matrix
	D3DXMatrixLookAtLH
	(
		&matCamView,
		&vecCamPos,
		&vecCamLookAt,
		&vecCamUp
	);

	float screenaspect = v(screen_size).x/v(screen_size).y;
	float fovy = 2.0 * atan(tan(_FLOAT(CamView->arc))/screenaspect);

	//calculate the cameras projection matrix
	D3DXMatrixPerspectiveFovLH
	(
		&matCamProj,
		fovy,
		screenaspect,
		_FLOAT(clipnear),
		_FLOAT(clipfar)
	);

	//calculate the cameras view projection matrix
	D3DXMatrixMultiply
	(
		&matCamViewProj,
		&matCamView,
		&matCamProj
	);

	//calculate the cameras inverse projection
	D3DXMatrixInverse
	(
		&matCamInvViewProj,
		NULL,
		&matCamViewProj
	);


	//get the ligth position
	vecLightPos = D3DXVECTOR3(_FLOAT(v(sun_pos).x), _FLOAT(v(sun_pos).z), _FLOAT(v(sun_pos).y));
	vecCamDir = vecCamLookAt-vecCamPos;
	D3DXVec3Normalize(&vecCamDir,&vecCamDir);
	vecLightDir = vecLightLookAt-vecLightPos;
	D3DXVec3Normalize(&vecLightDir,&vecLightDir);

	vecLightUp =
			D3DXVECTOR3
			(
				vecCamDir.x,
				vecCamDir.y,
				vecCamDir.z
			);

	//calculate the light view matrix
	D3DXMatrixLookAtLH
	(
		&matLightView,
		&vecLightPos,
		&vecCamPos,
		&vecLightUp
	);

	//set the viewcorners positions
	D3DXVECTOR4 vecViewCornerPos[8];
	vecViewCornerPos[0] = D3DXVECTOR4(-1,-1,-1,1);
	vecViewCornerPos[1] = D3DXVECTOR4(1,-1,-1,1);
	vecViewCornerPos[2] = D3DXVECTOR4(-1,1,-1,1);
	vecViewCornerPos[3] = D3DXVECTOR4(-1,-1,1,1);
	vecViewCornerPos[4] = D3DXVECTOR4(-1,1,1,1);
	vecViewCornerPos[5] = D3DXVECTOR4(1,1,-1,1);
	vecViewCornerPos[6] = D3DXVECTOR4(1,-1,1,1);
	vecViewCornerPos[7] = D3DXVECTOR4(1,1,1,1);

	//transform them to viewspace
	int i;
	for(i=0;i<8;i++)
	{
		D3DXVec4Transform
		(
			&vecViewCornerPos[i],
			&vecViewCornerPos[i],
			&matCamInvViewProj
		);

		D3DXVec4Scale
		(
			&vecViewCornerPos[i],
			&vecViewCornerPos[i],
			1/vecViewCornerPos[i].w
		);

		D3DXVec4Transform
		(
			&vecViewCornerPos[i],
			&vecViewCornerPos[i],
			&matLightView
		);
	}


	//get the min and max positions of the corners
	D3DXVECTOR3 vecMin = vecViewCornerPos[0];
	D3DXVECTOR3 vecMax = vecViewCornerPos[0];
	for(i=1;i<8;i++)
	{
		vecMin.x = min(vecMin.x,vecViewCornerPos[i].x);
		vecMax.x = max(vecMax.x,vecViewCornerPos[i].x);

		vecMin.y = min(vecMin.y,vecViewCornerPos[i].y);
		vecMax.y = max(vecMax.y,vecViewCornerPos[i].y);

		vecMin.z = min(vecMin.z,vecViewCornerPos[i].z);
		vecMax.z = max(vecMax.z,vecViewCornerPos[i].z);
	}

	//calculate the lights projection matrix
	D3DXMatrixOrthoOffCenterLH
	(
		&matLightProj,
		vecMin.x,
		vecMax.x,
		vecMin.y,
		vecMax.y,
		vecMin.z-10000,
		vecMax.z+10000
	);

	//calculate the lights view projection matrix
	D3DXMatrixMultiply
	(
		&matLightViewProj,
		&matLightView,
		&matLightProj
	);

	//pass the new viewprojectionmatrix to mtl.matrix
	ENGINE_MATERIAL Material = ev->mtl;
	Material->matrix[0] = matLightViewProj._11;
	Material->matrix[1] = matLightViewProj._12;
	Material->matrix[2] = matLightViewProj._13;
	Material->matrix[3] = matLightViewProj._14;
	Material->matrix[4] = matLightViewProj._21;
	Material->matrix[5] = matLightViewProj._22;
	Material->matrix[6] = matLightViewProj._23;
	Material->matrix[7] = matLightViewProj._24;
	Material->matrix[8] = matLightViewProj._31;
	Material->matrix[9] = matLightViewProj._32;
	Material->matrix[10] = matLightViewProj._33;
	Material->matrix[11] = matLightViewProj._34;
	Material->matrix[12] = matLightViewProj._41;
	Material->matrix[13] = matLightViewProj._42;
	Material->matrix[14] = matLightViewProj._43;
	Material->matrix[15] = matLightViewProj._44;

	ev->matView[0] = matLightView._11;
	ev->matView[1] = matLightView._12;
	ev->matView[2] = matLightView._13;
	ev->matView[3] = matLightView._14;
	ev->matView[4] = matLightView._21;
	ev->matView[5] = matLightView._22;
	ev->matView[6] = matLightView._23;
	ev->matView[7] = matLightView._24;
	ev->matView[8] = matLightView._31;
	ev->matView[9] = matLightView._32;
	ev->matView[10] = matLightView._33;
	ev->matView[11] = matLightView._34;
	ev->matView[12] = matLightView._41;
	ev->matView[13] = matLightView._42;
	ev->matView[14] = matLightView._43;
	ev->matView[15] = matLightView._44;

	ev->matProj[0] = matLightProj._11;
	ev->matProj[1] = matLightProj._12;
	ev->matProj[2] = matLightProj._13;
	ev->matProj[3] = matLightProj._14;
	ev->matProj[4] = matLightProj._21;
	ev->matProj[5] = matLightProj._22;
	ev->matProj[6] = matLightProj._23;
	ev->matProj[7] = matLightProj._24;
	ev->matProj[8] = matLightProj._31;
	ev->matProj[9] = matLightProj._32;
	ev->matProj[10] = matLightProj._33;
	ev->matProj[11] = matLightProj._34;
	ev->matProj[12] = matLightProj._41;
	ev->matProj[13] = matLightProj._42;
	ev->matProj[14] = matLightProj._43;
	ev->matProj[15] = matLightProj._44;
}
